/////////////////////////////////////////////////////////////////////////////
// File     : x16demo.c                                                    //
// Function : QNX demo application for Sensoray model 516/616 boards       //
// Author   : Jim Lamberson and Charlie Liu                                //
// Revision :                                                              //
//            Sept, 2003   Jim Lamberson    release sx16-0.7 for Linux     //
//            Oct., 2003   Charlie X. Liu   Added caption support          //
//            Nov., 2003   Charlie X. Liu   Ported/migrated to QNX         //
//            Jan., 2004   Charlie X. Liu   Added optional caption feature //
//                                          -- overlaying with system time //
/////////////////////////////////////////////////////////////////////////////

/*
Copyright (C) 2003-2004 Sensoray Co., Inc.  This program is a demo software;

The program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You must obey the GNU General Public License in all respects for all of
the code used. If you modify this file, you may extend this exception
to your version of the file, but you are not obligated to do so. If you
do not wish to do so, delete this exception statement from your version.
*/

// The necessary headers ///////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <sched.h>
#include <sys/ioctl.h>

#include "sx16api.h"			// sx16 driver api.
#include "sx16.h"			    // ioctl definitions for sx16 driver.
#include "sx16core.h"           // x16 attributes.


// Manifest constants //////////////////////////////////////////////////////////////////////

#define ENABLED				1
#define DISABLED			0

#define TRUE				1
#define FALSE				0

#define MAXLEN				80			// Maximum length of user command string.
#define MAXNUMBOARDS		4			// Maximum number of boards supported by the driver.

// Stream mode:
#define MODE_IDLE			0			// Halted.
#define MODE_PLAY			1			// Streaming from file.
#define MODE_RECORD			2			// Streaming to file.


// Configuration constants /////////////////////////////////////////////////////////////////

#define NUM_BOARDS		1				// Number of x16 boards in the system (1 to 4).
#define MPEG_BUF_SIZE	4096			// Number of MPEG bytes to transfer in a single chunk.  Max size is 32K.
#define BLOCKING_MODE	TRUE			// Set to TRUE to use blocking read/write operations.

// Device names for the x16 boards. 
#define DEV_NAME_0		"/dev/sx16a0"
#define DEV_NAME_1		"/dev/sx16a1"
#define DEV_NAME_2		"/dev/sx16a2"
#define DEV_NAME_3		"/dev/sx16a3"


// Storage //////////////////////////////////////////////////////////////////

typedef struct BOARD {					// BOARD OBJECT -------
  int					handle;			//  Handle to x16 device.
  int					mode;			//  Stream mode.
  int					stop_rq;		//  Stream Halt requested.
  char					fname[MAXLEN];	//  File name for mpeg data.
  pthread_t				io_thread_id;	//  Handle to stream i/o thread.
  pthread_t				systm_thrd_id; //  Handle to overlaying system clock thread
  unsigned int			mpeg_count;		//  Number of MPEG bytes transferred.
  int					flags;			//  Status flags from i/o thread to main thread.
} BOARD;

static struct	BOARD bd[NUM_BOARDS];	// Board objects.
static int		act_bdnum	= 0;		// Board 0 is active by default.


// Function prototypes ///////////////////////////////////////////////////////////////////////////////

static int  app_init( void );
static void app_shutdown( void );
static void cmd_processor( void );
static void *io_thread_func( void* arg );
static void get_attributes( struct BOARD *b );
static void set_attribute( struct BOARD *b, int val, int minval, int maxval, int ioctl_num );
static void get_caption_settings( struct BOARD *b );
static void start_stream( struct BOARD *b, char *fname, int new_stream_mode );
static void stop_stream( struct BOARD *b );
static void	*osystm_thread_func( void *arg );
static void overlay_systm( struct BOARD *b, int on );


///////////////////////////////////////////////////////////////////////////
// Init the system and execute user commands until termination.

int main( int argc, char *argv[] )
{
  int rtnval;

  // Attempt to open and initialize all boards.
  rtnval = app_init();

  // If all boards successfully opened and initialized ...
  if ( rtnval == 0 )
  {
    // Execute user commands until Application Quit command is invoked.
    cmd_processor();
  }

  // Close all opened boards.
  app_shutdown();

  // Exit application.
  return 0;
}

///////////////////////////////////////////////////////////////////////
// Open and initialize one board.  Returns: 0=success, -1=error.

static int board_open( int board_num, char *dev_name )
{
  // Set pointer to board's object.
  struct BOARD *b = &bd[board_num];

  // Initialize object.
  b->mode			= MODE_IDLE;
  b->fname[0]	= 0;
  b->flags		= 0;

  // Open device.
  b->handle = sx16_open( );
  if ( b->handle == -1 )
  {
    printf( "Driver open error\n" );
    return( -1 );
  }

  // Program the default settings.
  sx16_ioctl ( b->handle, SX16_IOC_SET_COLORBARS,		DISABLED	);
  sx16_ioctl ( b->handle, SX16_IOC_SET_VIDEO_SYSTEM,	SYS_NTSC	);
  sx16_ioctl ( b->handle, SX16_IOC_SET_BITRATE,			1000000		);
  sx16_ioctl ( b->handle, SX16_IOC_SET_PICSIZE,			SIZE352_2	);
  sx16_ioctl ( b->handle, SX16_IOC_SET_INPUT,			0			);
  sx16_ioctl ( b->handle, SX16_IOC_SET_INPUT_TYPE,		CVIDEO		);
  sx16_ioctl ( b->handle, SX16_IOC_SET_VBR,				DISABLED	);
  sx16_ioctl ( b->handle, SX16_IOC_SET_AUDIO_VOLUME,	6			);
  sx16_ioctl ( b->handle, SX16_IOC_SET_AUDIO_MUTE,		DISABLED	);

  // Set the caption defaults.
  sx16_ioctl ( b->handle, SX16_IOC_SET_CAPTION_WINDOW,			1				);	// window #1
  sx16_ioctl ( b->handle, SX16_IOC_SET_CAPTION_XSTART,			0				);	// 0
  sx16_ioctl ( b->handle, SX16_IOC_SET_CAPTION_YSTART,			0				);	// 0
  sx16_ioctl ( b->handle, SX16_IOC_SET_CAPTION_SCALE,			OVL_SCALE_X1	);	// x1
  sx16_ioctl ( b->handle, SX16_IOC_SET_CAPTION_COLOR,			OVL_WHITE		);	// white
  sx16_ioctl ( b->handle, SX16_IOC_SET_CAPTION_TRANSPARENCY,	OVL_TRANSP_T	);	// transparent
  sx16_ioctl ( b->handle, SX16_IOC_SET_CAPTION_ON,				CAPTION_OFF		);	// off

}

///////////////////////////////////////////////////////////////////////
// Open and initialize all boards.  Returns: -1=error, 0=success.

static int app_init( void )
{
  int		i;
  char	dev_name[80];

  // For each board ...
  for ( i = 0; i < NUM_BOARDS; i++ )
  {
    // Specify board's device name.
    switch ( i )
    {
    case 0:	strcpy( dev_name, DEV_NAME_0 );	break;
    case 1:	strcpy( dev_name, DEV_NAME_1 );	break;
    case 2:	strcpy( dev_name, DEV_NAME_2 );	break;
    case 3:	strcpy( dev_name, DEV_NAME_3 );	break;
    }

    // Open device.  Return error if open failed.
    if ( board_open( i, dev_name ) == -1 )
      return -1;
  }

  return 0;
}

//////////////////////////////////////////////////////////////
// Terminate all stream threads and close all boards.

static void app_shutdown( void )
{
  struct BOARD *b;

  // For each board ...
  for ( b = bd; b < &bd[NUM_BOARDS]; b++ )
  {
    // We are finished if this board was never opened.
    if ( b->handle == -1 )
      break;

    // Terminate the board's I/O thread if it exists.
    stop_stream( b );

    // Close the board.
    sx16_release( );	// close().
  }

  printf( "\n" );
}

//////////////////////////////////////////////////////////////
// Display the user prompt.

static void prompt( void )
{
  printf( "x16demo[%d]>", act_bdnum );
  fflush( stdout );
}

//////////////////////////////////////////////////////////////
// Display the command menu.

static void show_menu( void )
{
  printf( "+-SX16DEMO-COMMANDS------------------------------------------------------------+\n" );
  printf( "|   board n ....... select board: 1..4                                         |\n" );
  printf( "|   chan n ........ set video channel: 0 to 3                                  |\n" );
  printf( "|   get ........... display current settings                                   |\n" );
  printf( "|   intype n ...... set video type: 0=S-Video, 1=Composite                     |\n" );
  printf( "|   loopback ...... set video codec loopback (516 only): 0|1                   |\n" );
  printf( "|   rgpio n ....... display GPIO output states                                 |\n" );
  printf( "|   wgpio n ....... set GPIO output states: 0=off, 1=on                        |\n" );
  printf( "|   h[elp] ........ display this list                                          |\n" );
  printf( "|   q[uit] ........ quit program                                               |\n" );
  printf( "|   reset ......... reset board                                                |\n" );
  printf( "+-AUDIO------------------------------------------------------------------------+\n" );
  printf( "|   arate n ....... set bits/second:                                           |\n" );
  printf( "|                     1=32K, 2=48K, 3=56K, 4=64K, 5=80K, 6=96K, 7=112K, 8=128K |\n" );
  printf( "|                     9=160K, 10=192K, 11=224K, 12=256K, 13=320K, 14=384K      |\n" );
  printf( "|   aformat n ..... set format:                                                |\n" );
  printf( "|                     0=mpeg1/layer1, 1=mpeg1/layer2, 2=mp3, 3=ac3/2-ch        |\n" );
  printf( "|   stereo n ...... set stereo mode:                                           |\n" );
  printf( "|                     mpeg: 0=stereo, 1=joint, 2=2-ch, 3=1-ch                  |\n" );
  printf( "|                     ac3:  0=1-ch, 1=2-ch, 2=undef, 3=undef                   |\n" );
  printf( "|   mute n ........ set mute enable: 0=no mute, 1=mute                         |\n" );
  printf( "|   vol n ......... set volume: 0..63 (-dB units)                              |\n" );
  printf( "+-VIDEO------------------------------------------------------------------------+\n" );
  printf( "|   bars n ........ set output colorbars enable: 0=off, 1=on                   |\n" );
  printf( "|   bright n ...... set luminance brightness: 0..255                           |\n" );
  printf( "|   contrast n .... set luminance contrast: 0..255                             |\n" );
  printf( "|   vsat n ........ set chrominance saturation: 0..255                         |\n" );
  printf( "|   vsystem n ..... set video system: 1=PAL, 2=NTSC                            |\n" );
  printf( "+-MPEG-------------------------------------------------------------------------+\n" );
  printf( "|   gop n ......... set GOP: 0 to 31                                           |\n" );
  printf( "|   ipdist n ...... set IP distance: 0 to 3                                    |\n" );
  printf( "|   mpeg n ........ set mpeg type and picture size:                            |\n" );
  printf( "|                     0=mpeg2:D1-704, 1=mpeg2:2/3 D1, 2=mpeg2:1/2 D1,          |\n" );
  printf( "|                     3=mpeg1:SIF,    4=mpeg1:QSIF,                            |\n" );
  printf( "|                     5=mpeg2:D1-720, 6=mpeg2:D1-640                           |\n" );
  printf( "|   stream n ...... set stream type: 0=mux, 1=video es, 2=audio es             |\n" );
  printf( "|   avsel n ....... set mux a/v select: 0=aud+vid, 1=vid only, 2=aud only      |\n" );
  printf( "|   vrate n ....... set video bits/second: 100000..10000000                    |\n" );
  printf( "|   vbr n ......... set variable bit rate enable: 0=off, 1=on                  |\n" );
  printf( "+-STREAM-----------------------------------------------------------------------+\n" );
  printf( "|   play fname .... play from MPEG file; default name is \"tmp.mpg\"             |\n" );
  printf( "|   rec fname ..... record to MPEG file; default name is \"tmp.mpg\"             |\n" );
  printf( "|   stop .......... stop MPEG stream and close file                            |\n" );
  printf( "+-CAPTION----------------------------------------------------------------------+\n" );
  printf( "|   window n ...... set window number: 1 or 2                                  |\n" );
  printf( "|   xstart n ...... set window position for x_start: 8..700                    |\n" );
  printf( "|   ystart n ...... set window position for y_start: 0..478                    |\n" );
  printf( "|   scale n ....... set scale for font:                                        |\n" );
  printf( "|                     0=x1, 1=x2, 2=x2(only vertical), 3=x4                    |\n" );
  printf( "|   color n ....... set color for caption:                                     |\n" );
  printf( "|                     0=white, 1=black,                                        |\n" );
  printf( "|   transp n ...... set transparency attribute for caption:                    |\n" );
  printf( "|                     1=transparent,     2=black-background                    |\n" );
  printf( "|                     3=transp-text on white-bkgrnd                            |\n" );
  printf( "|                     4=black-text on white-bkgrnd                             |\n" );
  printf( "|   text string ... set text string for overlay:  any ascii string             |\n" );
  printf( "|   caption n ..... set caption off/on: 0=off, 1=on                            |\n" );
  printf( "|   osystm n ...... set current caption window overlaying with system time:    |\n" );
  printf( "|                     0=off, 1=on                                              |\n" );
  printf( "+------------------------------------------------------------------------------+\n" );
}

///////////////////////////////////////////////////////////////////////////////
// Decode and execute user command.  Returns TRUE if Quit selected.

static int exec_command( char *cmd, char *arg )
{
  struct BOARD	*b;

  // Convert argument to integer value (not relevant for all commands).
  int arg_val = -1;
  sscanf( arg, "%d", &arg_val );

  // Decode and execute command -----------------------------------

  // Terminate application.
  if ( cmd[0] == 'q' )
    return TRUE;

  // Select board.
  if	( strcmp( cmd, "board" ) == 0 )
  {
    if ( ( arg_val >= 0 ) && ( arg_val <= NUM_BOARDS ) )
      act_bdnum = arg_val;
    else
      printf( "Illegal or missing argument; must be in range 0..%d\n", NUM_BOARDS );
    prompt();
    return FALSE;
  }

  // Show help.
  if ( ( cmd[0] == 'h' ) || ( cmd[0] == '?' ) )
  {
    show_menu();
    prompt();
    return FALSE;
  }

  // Prompt for user command.
  prompt();

  // Set pointer to active board's object.
  b = &bd[act_bdnum];

  // Set video attribute.
  if      ( strcmp( cmd, "bright"    ) == 0 )	set_attribute( b, arg_val, 0,	              255,        			SX16_IOC_SET_BRIGHTNESS    );
  else if ( strcmp( cmd, "intype"    ) == 0 )	set_attribute( b, arg_val, CVIDEO,	          SVIDEO,    			SX16_IOC_SET_INPUT_TYPE    );
  else if ( strcmp( cmd, "chan"      ) == 0 )	set_attribute( b, arg_val, 0,				  3,          			SX16_IOC_SET_INPUT		   );
  else if ( strcmp( cmd, "bars"      ) == 0 )	set_attribute( b, arg_val, DISABLED,	      ENABLED,    			SX16_IOC_SET_COLORBARS	   );
  else if ( strcmp( cmd, "contrast"  ) == 0 )	set_attribute( b, arg_val, 0,	              255,		    		SX16_IOC_SET_CONTRAST      );
  else if ( strcmp( cmd, "vsystem"   ) == 0 )	set_attribute( b, arg_val, SYS_PAL,		      SYS_NTSC,   			SX16_IOC_SET_VIDEO_SYSTEM  );

	// Set mpeg codec attribute.
  else if ( strcmp( cmd, "mpeg"      ) == 0 )	set_attribute( b, arg_val, SIZE704_2,	      SIZE640_2,  			SX16_IOC_SET_MPEG		   );
  else if ( strcmp( cmd, "stream"    ) == 0 )	set_attribute( b, arg_val, STREAMTYPE_MUX,    STREAMTYPE_AES,		SX16_IOC_SET_STREAM_TYPE   );
  else if ( strcmp( cmd, "vrate"     ) == 0 )	set_attribute( b, arg_val, 100000,		      20000000,   			SX16_IOC_SET_BITRATE	   );
  else if ( strcmp( cmd, "vbr"       ) == 0 )	set_attribute( b, arg_val, DISABLED,	      ENABLED,    			SX16_IOC_SET_VBR		   );
  else if ( strcmp( cmd, "vsat"      ) == 0 )	set_attribute( b, arg_val, 0,	              255,		    		SX16_IOC_SET_SATURATION    );
  else if ( strcmp( cmd, "avsel"     ) == 0 )	set_attribute( b, arg_val, 0,	              2,	  		  		SX16_IOC_SET_AV_SELECT	   );
  else if ( strcmp( cmd, "prognum"   ) == 0 )	set_attribute( b, arg_val, 0,	              255,  		  		SX16_IOC_SET_PROG_NUM	   );  // not in menu
  else if ( strcmp( cmd, "gop"       ) == 0 )	set_attribute( b, arg_val, 0,	              31,	  	    		SX16_IOC_SET_M             );
  else if ( strcmp( cmd, "ipdist"    ) == 0 )	set_attribute( b, arg_val, 0,	              3,	  	    		SX16_IOC_SET_N             );

  // Set audio attribute.
  else if ( strcmp( cmd, "vol"       ) == 0 )	set_attribute( b, arg_val, 0,				  63,					SX16_IOC_SET_AUDIO_VOLUME  );
  else if ( strcmp( cmd, "mute"      ) == 0 )	set_attribute( b, arg_val, DISABLED,	      ENABLED,				SX16_IOC_SET_AUDIO_MUTE    );
  else if ( strcmp( cmd, "arate"     ) == 0 )	set_attribute( b, arg_val, 1,     		      14,        			SX16_IOC_SET_AUDIO_BITRATE );
  else if ( strcmp( cmd, "aformat"   ) == 0 )	set_attribute( b, arg_val, AFMT_MPEG1_LAYER1, AFMT_AC3,				SX16_IOC_SET_AUDIO_FORMAT  );
  else if ( strcmp( cmd, "stereo"    ) == 0 )	set_attribute( b, arg_val, STEREOMODE_STEREO, STEREOMODE_SINGLE,	SX16_IOC_SET_AUDIO_STEREO  );

  // Set/get GPIO output states.
  else if ( strcmp( cmd, "wgpio"     ) == 0 ) sx16_ioctl( b->handle, SX16_IOC_SET_GPIO, arg_val );
  else if ( strcmp( cmd, "rgpio"     ) == 0 ) printf( "gpio=%d\n", sx16_ioctl( b->handle, SX16_IOC_GET_GPIO, arg_val ) );

	// Display current settings.
  else if ( strcmp( cmd, "get"       ) == 0 )	{ get_attributes( b ); get_caption_settings( b); prompt(); }

  // Control MPEG stream mode.
  else if ( strcmp( cmd, "rec"       ) == 0 )	start_stream( b, ( arg[0] == 0 ) ? "tmp.mpg" : arg, MODE_RECORD );
  else if ( strcmp( cmd, "play"      ) == 0 )	start_stream( b, ( arg[0] == 0 ) ? "tmp.mpg" : arg, MODE_PLAY   );
  else if ( strcmp( cmd, "stop"      ) == 0 )	stop_stream( b );

  // Set Caption attributes/parameters.
  else if ( strcmp( cmd, "window"    ) == 0 )	set_attribute( b, arg_val, 1,				2,					SX16_IOC_SET_CAPTION_WINDOW );
  else if ( strcmp( cmd, "xstart"    ) == 0 )	set_attribute( b, arg_val, 0,				MAX_XSTART,			SX16_IOC_SET_CAPTION_XSTART );
  else if ( strcmp( cmd, "ystart"    ) == 0 )	set_attribute( b, arg_val, 0,				MAX_YSTART,			SX16_IOC_SET_CAPTION_YSTART );
  else if ( strcmp( cmd, "scale"     ) == 0 )	set_attribute( b, arg_val, OVL_SCALE_X1,	OVL_SCALE_X4,		SX16_IOC_SET_CAPTION_SCALE );
  else if ( strcmp( cmd, "color"     ) == 0 )	set_attribute( b, arg_val, OVL_WHITE,		OVL_BLACK,			SX16_IOC_SET_CAPTION_COLOR );
  else if ( strcmp( cmd, "transp"    ) == 0 )	set_attribute( b, arg_val, OVL_TRANSP_T,	OVL_TRANSP_B_BKGRD,	SX16_IOC_SET_CAPTION_TRANSPARENCY );
  else if ( strcmp( cmd, "text"      ) == 0 )	sx16_ioctl( b->handle, SX16_IOC_SET_CAPTION_TEXT, (unsigned long) arg );
  else if ( strcmp( cmd, "caption"   ) == 0 )	set_attribute( b, arg_val, CAPTION_OFF,		CAPTION_ON,			SX16_IOC_SET_CAPTION_ON );
  else if ( strcmp( cmd, "osystm"    ) == 0 )	overlay_systm( b, ( arg_val == 1 ) ? 1 : 0 );

	// Other commands.
  else if ( strcmp( cmd, "loopback"  ) == 0 )	sx16_ioctl( b->handle, SX16_IOC_GA_LOOPBACK, arg_val );
  else if ( strcmp( cmd, "reset"     ) == 0 )	sx16_ioctl( b->handle, SX16_IOC_RESET, 0 );

	// Depracated commands -- NOT RECOMMENDED FOR NEW DESIGNS.
  else if ( strcmp( cmd, "muxformat" ) == 0 )	set_attribute( b, arg_val, 0,	             6,	        		SX16_IOC_SET_MUX_FORMAT    );
  else if ( strcmp( cmd, "picsize"   ) == 0 )	set_attribute( b, arg_val, SIZE704_2,	     SIZE640_2,  		SX16_IOC_SET_PICSIZE	   );
  else if ( strcmp( cmd, "sysformat" ) == 0 )	set_attribute( b, arg_val, 0,	             2,	  	    		SX16_IOC_SET_SYS_FORMAT    );

  // Illegal command.
  else if ( cmd[0] != 0 )
  {
    printf( "Invalid command: %s\n", cmd );
    prompt();
  }

  return FALSE;
}

////////////////////////////////////////////////////////////////////
// Execute user commands until Application Quit is requested.

static void cmd_processor( void )
{
  char					cmdbuf[MAXLEN];
  char					cmd[MAXLEN];
  char					arg[MAXLEN];
  char					*c;
  int					arg_val;

  // Display the command menu.
  show_menu();

  // Prompt for user command.
  prompt();

  // Repeat until Application Terminate is requested ...
  for ( ; ; )
  {
    // Input user command string, and if no errors encountered ...
    if ( fgets( cmdbuf, MAXLEN - 1, stdin ) != NULL )
    {
      // Truncate newline char at end of command string.
      cmdbuf[ strlen( cmdbuf ) - 1 ] = 0;

      // If command line is not blank ...
      if ( cmdbuf[0] != 0 )
      {
        // Force command string to lower case, except command=="text".
		if ( strncmp( cmdbuf, "text", 4 ) != 0 )
			for ( c = cmdbuf; *c != 0; c++ ) 
				*c = tolower( *c );

        // Parse command string into opcode and argument.
        cmd[0] = arg[0] = 0;
        sscanf( cmdbuf, "%s %s", cmd, arg );

        // Decode and execute command
        if ( exec_command( cmd, arg ) )
          break;
      }
      else
      {
        // Prompt for user command.
        prompt();
      }
    }
  }
}

///////////////////////////////////////////////////
// Set programmable attribute on a board.

static void set_attribute( struct BOARD *b, int val, int minval, int maxval, int ioctl_num )
{
  // Set attribute if value is in valid range, otherwise display error message.
  if ( ( val >= minval ) && ( val <= maxval ) )
    sx16_ioctl( b->handle, ioctl_num, val );
  else
  {
    printf( "Illegal or missing argument; must be in range %d..%d\n", minval, maxval );
    prompt();
  }
}

/////////////////////////////////////////////////////
// Fetch and display board's current settings.

static char txt_muxformat	[][40] = { "svcd", "vcd", "dvd", "cdda", "video-es", "audio-es", "transport" };
static char txt_arate		[][40] = {	"", "32", "48", "56", "64", "80", "96", "112", "128", "160", "192", "224", "256", "320", "384" };
static char txt_mpeg		[][40] = { "mpeg2:D1-704", "mpeg2:2/3-D1", "mpeg2:1/2-D1", "mpeg1:SIF", "mpeg1:QSIF", "mpeg2:D1-720", "mpeg2:D1-640" };
static char txt_boolean		[][40] = { "off", "on" };
static char txt_vsystem		[][40] = { "", "pal", "ntsc" };
static char txt_intype		[][40] = { "s-video", "composite" };
static char txt_sysformat	[][40] = { "svcd", "vcd", "other" };
static char txt_avsel		[][40] = { "audio+video", "video only", "audio only" };
static char txt_stereomode	[][40] = { "stereo", "joint stereo", "dual channel", "single channel" };
static char txt_asmprate	[][40] = { "44", "48", "32" };
static char txt_pcmres		[][40] = { "16", "18", "20", "24" };
static char txt_aformat		[][40] = { "mpeg1:layer-1", "mpeg1:layer-2", "mp3", "ac3:2-ch" };
static char txt_cap_scale	[][40] = { "x1", "x2", "x2 (vertical only)", "x4" };
static char txt_cap_color	[][40] = { "white", "black", "red", "green", "blue",  "yellow", "cyan", "magenta" };
static char txt_cap_transp	[][40] = { "", "transparent", "black-background", "transp-text on white-bkgrnd", "black-text on white-bkgrnd" }; 
static char txt_cap_on		[][40] = { "off", "on" };

// Fetch and display the attributes.
static void get_attributes( struct BOARD *b )
{
	printf( "\nCURRENT SETTINGS -----------\n" );
	printf( "audio bass           = %d\n", sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_bass				) );
	printf( "audio treble         = %d\n", sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_treble			) );
	printf( "audio volume         = %d\n", sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_volume			) );
	printf( "stereo mode          = %d\n", sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_stereo_mode		) );
	printf( "video bitrate (bps)  = %d\n", sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_bitrate			) );
	printf( "video brightness     = %d\n", sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_brightness		) );
	printf( "video channel        = %d\n", sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_chan				) );
	printf( "video contrast       = %d\n", sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_contrast			) );
	printf( "gop                  = %d\n", sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_m				) );
	printf( "ip dist              = %d\n", sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_n				) );
	printf( "video saturation     = %d\n", sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_saturation		) );
	printf( "program number       = %d\n", sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_program_number			) );

	printf( "audio format         = %s\n", txt_aformat		[ sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_format				) ] );
	printf( "a/v select           = %s\n", txt_avsel		[ sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_av_select					) ] );
	printf( "audio stereo mode    = %s\n", txt_stereomode	[ sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_stereo_mode			) ] );
	printf( "audio samprate (KHz) = %s\n", txt_asmprate		[ sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_sample_rate			) ] );
	printf( "audio pcm res (bits) = %s\n", txt_pcmres		[ sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_pcm_resolution		) ] );
	printf( "video type           = %s\n", txt_intype		[ sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_type					) ] );
	printf( "audio bitrate (kbps) = %s\n", txt_arate		[ sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_audio_bitrate				) ] );
	printf( "video colorbars      = %s\n", txt_boolean		[ sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_colorbars_enable		) ] );
	printf( "video filter         = %s\n", txt_boolean		[ sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_filter_enable			) ] );
	printf( "video system         = %s\n", txt_vsystem		[ sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_system				) ] );
	printf( "variable bitrate     = %s\n", txt_boolean		[ sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_vbr_enable			) ] );
	printf( "system format        = %s\n", txt_sysformat	[ sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_system_format			) ] );
	printf( "mpeg type/picsize    = %s\n", txt_mpeg			[ sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_video_picsize				) ] );
	printf( "mux format           = %s\n", txt_muxformat	[ sx16_ioctl( b->handle, SX16_IOC_GET_SETTING, ATTR_mux_format					) ] );
}

// Fetch and display the caption settings.
static void get_caption_settings( struct BOARD *b )
{
	printf( "caption window       = %d\n", sx16_ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_window		) );
	printf( "caption X start      = %d\n", sx16_ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_xstart		) );
	printf( "caption X stop       = %d\n", sx16_ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_xstop		) );
	printf( "caption Y start      = %d\n", sx16_ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_ystart		) );
	printf( "caption Y stop       = %d\n", sx16_ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_ystop		) );
	printf( "caption scale        = %s\n", txt_cap_scale	[sx16_ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_scale		) ] );
	printf( "caption color        = %s\n", txt_cap_color	[sx16_ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_color		) ] );
	printf( "caption transparency = %s\n", txt_cap_transp	[sx16_ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_transp		) ] );
	printf( "caption on/off       = %s\n", txt_cap_on		[sx16_ioctl( b->handle, SX16_IOC_GET_CAPTION_SETTINGS, CAPTN_on			) ] );
}

///////////////////////////////////////////////////
// Start MPEG data streaming to/from a file.

static void start_stream( struct BOARD *b, char *fname, int new_stream_mode )
{
  pthread_attr_t attr;

  // If we are already streaming ...
  if ( b->mode != MODE_IDLE )
  {
    // Abort command.
    printf( "already streaming\n" );
    prompt();
    return;
  }

  // Make the MPEG filename visible to the I/O thread.
  strcpy( b->fname, fname );

  // Indicate the new stream mode.
  b->mode = new_stream_mode;

  // Clear the stream thread's termination request.
  b->stop_rq = 0;

  // Create detached thread that will manage the MPEG stream flow.
  pthread_attr_init( &attr );
  pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_DETACHED );
  pthread_create( &b->io_thread_id, &attr, &io_thread_func, b );
  pthread_attr_destroy( &attr );
}

//////////////////////////////////////////////////
// Halt MPEG data stream to/from a file.

static void stop_stream( struct BOARD *b )
{
  // If stream i/o thread is running ...
  if ( b->mode != MODE_IDLE )
  {
    // Request thread termination.
    b->stop_rq = 1;

    // Wait for the thread to terminate
    while ( b->mode != MODE_IDLE )
      sleep( 1 );
  }
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////  MPEG STREAM I/O THREAD  //////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////
// Copy a block of MPEG data from board to file.
// Returns number of bytes copied.

static int mpeg_to_file( struct BOARD *b, FILE *file )
{
  char	mpeg_buf[MPEG_BUF_SIZE];
  int		len;

  // Attempt to read MPEG data from board.
  len = sx16_read( b->handle, mpeg_buf, MPEG_BUF_SIZE );

  // Request Stream Halt and abort if we timed out while waiting to read MPEG data from board.
  if ( len <= 0 )
  {
    printf( "No signal\n" );
    prompt();
    b->stop_rq = 1;
    return 0;
  }

  // Write MPEG data to file.
  fwrite( mpeg_buf, 1, len, file );

  // Return number of bytes transferred.
  return len;
}

/////////////////////////////////////////////////////
// Copy a block of MPEG data from file to board.
// Returns number of bytes copied.

static int mpeg_from_file( struct BOARD *b, FILE *file )
{
  char	mpeg_buf[MPEG_BUF_SIZE];
  int	len;
  int	remlen;
  int	nwritten;

  // Attempt to read MPEG data from file.
  len = fread( mpeg_buf, 1, MPEG_BUF_SIZE, file );

  // Request Stream Halt and abort if end of file reached.
  if ( len == 0 )
  {
    printf( "End of file reached\n" );
    prompt();
    b->stop_rq = 1;
    return 0;
  }

  // Repeat until all MPEG bytes have been transferred to the board ...
  for ( remlen = len; remlen > 0; )
  {
    // Attempt to write MPEG data to board.
    nwritten = sx16_write( b->handle, mpeg_buf, remlen );

    //  Request Stream Halt and abort if we timed out while writing MPEG data to board.
    if ( nwritten < 0 )
    {
      printf( "Player error\n" );
      prompt();
      b->stop_rq = 1;
      return 0;
    }

    // Compute remaining byte count.
    remlen -= nwritten;
  }

  // Return number of bytes transferred.
  return len;
}

////////////////////////////////////////////////////////////////////
// I/O THREAD MAIN.  Read/write MPEG data until stream is stopped.
// The arg is a pointer to the device's BOARD object.

void *io_thread_func( void *arg )
{
  FILE					*file;					// MPEG data file.
  unsigned int			rlen;					// MPEG bytes transferred.
  time_t				starttime;				// Stream start time.
  time_t				stoptime;				// Stream stop time.
  BOARD					*b = (BOARD *)arg;		// Cache pointer to board object.

  // Select the function that will be used to move MPEG data between file and board.
  int (* mpeg_transfer)( struct BOARD *b, FILE *file ) = ( b->mode == MODE_RECORD ) ? mpeg_to_file : mpeg_from_file;

  // Initialize stream statistics.
//testing bgn
  usleep(10000);			// sleep for 10 ms to wait for stabalizing
//testing end
  b->mpeg_count = 0;		// No MPEG bytes have been transferred.
  time( &starttime );		// Log the start time.

  // Attempt to open the MPEG data file.  If successful ...
  if ( ( file = fopen( b->fname, ( b->mode == MODE_RECORD ) ? "wb" : "rb" ) ) != NULL )
  {
    // Report status.
    printf( "Starting stream -- %s", ctime(&starttime) );
    prompt();

    // Until stream termination is requested ...
    while ( !b->stop_rq )
    {
      // Transfer a block of MPEG data between file and board.  If this is the first
      // transfer, the driver will automatically start the stream running.
      rlen = mpeg_transfer( b, file );

      // Collect statistical info.
      b->mpeg_count += rlen;

      // Allow other processes to execute if we are in non-blocking mode and no data was available.
      if ( ( rlen == 0 ) && !BLOCKING_MODE )
      {
        // Sleep for approximately 10 milliseconds.
//        struct timespec tv;
//        tv.tv_sec = (time_t)0;
//        tv.tv_nsec = (long)( 0.01 * 1e+9 );
//        nanosleep( &tv, &tv );
		  usleep (10000);
      }
	  usleep(0);
    }

    // Halt the stream.
    sx16_ioctl( b->handle, SX16_IOC_RW_STOP, 0 );

    // Close the MPEG data file.
    fclose( file );

    // Report statistics.
    printf( "%u MPEG bytes transferred\n", b->mpeg_count );
    prompt();
  }

  else
  {
    // Indicate there was a problem opening the MPEG data file.
    printf( "file open error\n" );
    prompt();
  }

  // Indicate stream is stopped.
  b->mode = MODE_IDLE;

  // Report status.
  time( &stoptime );		// Log the stop time.
  printf( "Stream stopped  -- %s", ctime(&stoptime) );
  prompt();

  // Terminate thread.
  return 0;
}


////////////////////////////////////////////////////////////////////
// System time checking and overlaying thread.  Putting/overlaying
// with system time to current-operating caption window.

static void *osystm_thread_func( void *arg )
{
  BOARD			*b = (BOARD *) arg;		// Cache pointer to board object.
  time_t		time_now;				// time now.
  char			txt_time_now[30];
  static char	txt_time_last[30] = " \0";
  int			i;
  char			txt_osystm[30];

  while(1) {
	  time( &time_now );			// take out system time now.
	  strcpy ( txt_time_now, ctime(&time_now) );
	  if ( strcmp (txt_time_now, txt_time_last ) != 0 )
	  {
		  //printf ("*osysclk_thread_func(): %s", txt_time_now );
		  strcpy ( txt_osystm, txt_time_now );
		  for (i=0; i<30; i++) if (txt_osystm[i]=='\n') txt_osystm[i] = '\0';
		  sx16_ioctl( b->handle, SX16_IOC_SET_CAPTION_TEXT, (unsigned long) txt_osystm );
		  strcpy ( txt_time_last, txt_time_now );
	  }
	  usleep(100000);		// 0.1 second.
  }

}


//////////////////////////////////////////////////////////////
// Hanble creating&terminating overlaying system time thread.

static void overlay_systm( struct BOARD *b, int on )
{
	static int on_last = 0;
	pthread_attr_t attr;

	// Create detached thread that will overlay the system clock in the MPEG stream flow.
	if ( (on_last==0) && (on==1) ) {
		pthread_attr_init( &attr );
		pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_DETACHED );
		pthread_create( &b->systm_thrd_id, &attr, &osystm_thread_func, &b->handle );
		pthread_attr_destroy( &attr );
		on_last = on;
		return;
	}

	// Terminate the detached thread created by this function.
	if ( (on_last==1) && (on==0) ) {
		pthread_cancel( b->systm_thrd_id );
		on_last = on;
		return;
	}
	
}

